; misc utilities
; BP 200905

; helper functions

; vectors

(defun vec (x y z)
	(make-array 3 :initial-contents (list x y z))
)

(defun mulv (v1 v2) ;component-wise multiplication
	(let (
		(x1 (aref v1 0))
		(y1 (aref v1 1))
		(z1 (aref v1 2))
		(x2 (aref v2 0))
		(y2 (aref v2 1))
		(z2 (aref v2 2)))
		;in
		(vec (* x1 x2) (* y1 y2) (* z1 z2))
	)
)

(defun smulv (v1 sc) ;scalar multiplication
	(let (
		(x1 (aref v1 0))
		(y1 (aref v1 1))
		(z1 (aref v1 2)))
		;in
		(vec (* x1 sc) (* y1 sc) (* z1 sc))
	)
)

(defun subv (v1 v2) ;component-wise subtraction
	(let (
		(x1 (aref v1 0))
		(y1 (aref v1 1))
		(z1 (aref v1 2))
		(x2 (aref v2 0))
		(y2 (aref v2 1))
		(z2 (aref v2 2)))
		;in
		(vec (- x1 x2) (- y1 y2) (- z1 z2))
	)
)

(defun addv (v1 v2) ;component-wise subtraction
	(let (
		(x1 (aref v1 0))
		(y1 (aref v1 1))
		(z1 (aref v1 2))
		(x2 (aref v2 0))
		(y2 (aref v2 1))
		(z2 (aref v2 2)))
		;in
		(vec (+ x1 x2) (+ y1 y2) (+ z1 z2))
	)
)

(defun len (vec)
	(let ((x (aref vec 0)) (y (aref vec 1)) (z (aref vec 2)))
		(sqrt (+ (* x x) (* y y) (* z z)))
	)
)

(defun normalise (vec)
	(let* ((x (aref vec 0)) (y (aref vec 1)) (z (aref vec 2)) (l (sqrt (+ (* x x) (* y y) (* z z)))))
		(make-array 3 :initial-contents (list (/ x l) (/ y l) (/ z l)))
	)
)

(defun direction (v1 v2) ; return the direction vector from pos1 to po2
	(let (
		(x1 (aref v1 0))
		(y1 (aref v1 1))
		(z1 (aref v1 2))
		(x2 (aref v2 0))
		(y2 (aref v2 1))
		(z2 (aref v2 2)))
		; in
		(normalise (vec (- x2 x1) (- y2 y1) (- z2 z1)))
	)
)

(defun dot (v1 v2)
	(let (
		(x1 (aref v1 0))
		(y1 (aref v1 1))
		(z1 (aref v1 2))
		(x2 (aref v2 0))
		(y2 (aref v2 1))
		(z2 (aref v2 2)))
		
		;in
		(+ (* x1 x2) (* y1 y2) (* z1 z2))
	)
)

; colours

(defun colour (r g b)
	"Returns a colour object"
	(make-array 3 :initial-contents (list r g b))
)

(defun clamp (a b x)
	(cond
		((< x a) a)
		((> x b) b)
		(t x)
	)
)

(defun clamp-colour (col)
	"Clamps the colour to the proper range"
	(let ((r (aref col 0))(g (aref col 1))(b (aref col 2)))
		(colour (floor (clamp 0 255 r)) (floor (clamp 0 255 g)) (floor (clamp 0 255 b)))
	)
)

(defun addc (col1 col2) ; component-wise addition
	(let (
		(r1 (aref col1 0))
		(g1 (aref col1 1))
		(b1 (aref col1 2))
		(r2 (aref col2 0))
		(g2 (aref col2 1))
		(b2 (aref col2 2)))
		
		(colour (+ r1 r2) (+ g1 g2) (+ b1 b2))
	)
)

(defun mulc (col1 col2) ; component-wise multiplication (where 255*255 = 255)
	(let (
		(r1 (aref col1 0))
		(g1 (aref col1 1))
		(b1 (aref col1 2))
		(r2 (aref col2 0))
		(g2 (aref col2 1))
		(b2 (aref col2 2)))
		
		(colour (* 255 (/ r1 255.0) (/ r2 255.0)) (* 255 (/ g1 255.0) (/ g2 255.0)) (* 255 (/ b1 255.0) (/ b2 255.0)))
	)
)

(defun smulc (col1 sc) ; scalar multiplication (where colour*1 = colour)
	(let (
		(r1 (aref col1 0))
		(g1 (aref col1 1))
		(b1 (aref col1 2)))
		(colour (* r1 sc) (* g1 sc) (* b1 sc))
	)
)










	
