; A simple raytracer 
; BP 200905
; 
; Please, use it and abuse it.

(load "ppm")
(load "misc")

; info

; the camera is always at the origin and facing down the negative z-axis

; an object is a pair of functions (intersect,shader)
; intersect takes a ray (pos,dir) as a parameter and returns the closest point of intersection with that ray (if any)
; shader takes a position as a parameter and returns the colour of the surface at that point (it must do its own shading))

; constants

(defconstant *origin* (make-array 3 :initial-element 0))
(defconstant *max-bounces* 2)

; - colours

(defconstant *white* (colour 255 255 255))
(defconstant *black* (colour 0 0 0))
(defconstant *red* (colour 255 0 0))
(defconstant *blue* (colour 0 0 255))
(defconstant *green* (colour 0 255 0))
(defconstant *yellow* (colour 255 255 0))
(defconstant *orange* (colour 255 100 10))

; - background 

(defconstant *bg* *white*)
(defconstant *ambient* (colour 50 50 50))

; structures

(defstruct light
	(pos *origin*)
	(diffuse *white*)
	(specular *white*)
)

; globals

(setq object-list nil) ;fill this list with all your objects
(setq light-list nil)  ;fill this list with all your lights

; main render function

; default camera params
; proj plane is at z = 1
; width defaults to 1 unit across

(defun render (width height filename)
	"Renders an img from the default camera"		
	(let ((h (/ height width)) (img (new-img width height)))
		(dotimes (i height)
			(dotimes (j width)
				(let*
					(
						(y (- (/ h 2) (* (/ j height) h)))
						(x (- (/ i width) (/ 1 2)))
						(z -1)
						(pos (make-array 3 :initial-contents (list x y z)))
						(dir (normalise pos))
						; cast a ray from the camera thru the pixel at (i,j)
						(col (clamp-colour (cast-ray *origin* dir 0))) ;perspective proj
						;(col (cast-ray (make-array 3 :initial-contents (list x y 0)) (make-array 3 :initial-contents '(0 0 -1)) 0)) ;ortho proj
					)
					(set-pixel img i j (aref col 0) (aref col 1) (aref col 2))					
				)
			)
		)
		(img-to-file img filename)
	)		
)
; user interface
(defun intersect (pos dir)
	"Returns the closest intersection point and object, else nil"	
	(let ((closestPt nil) (closestObj nil))
		(dolist (object object-list)
			(let ((intersect (funcall (first object) pos dir)))
				(when intersect
					(when (or (not closestPt) (> (len (subv closestPt pos)) (len (subv intersect pos))))
						(setq closestPt intersect)
						(setq closestObj object)
					)
				)
			)
		)
		
		(if closestObj (list closestPt closestObj) nil)
	)
)

(defun cast-ray (pos dir bounce)
	"Returns a color triplet (array)"
	(let ((res (intersect pos dir)))
		(if res
			(let ((pt (first res)) (obj (second res)))
				(funcall (second obj) pt (1+ bounce))
			)
			*bg*
		)
	)		
)


