; a simple scene for the raytracer
; BP 200905

(compile-file "ray")
(load "ray.fas")

; a simple sphere
(let (
			(s-dark *red*) 
			(s-light *orange*)
			(s-ambient *white*) 
			(s-specular *white*) 
			(s-shininess 10) 
			(s-center (vec 0 0 -5))
			(s-radius 1)
			(s-reflection 0.6)
		)
	(setq sphere (list 
		;intersect function
		(lambda (pos dir)
			(let* (
				(Xd (aref dir 0))
				(Yd (aref dir 1))
				(Zd (aref dir 2))
				(X0 (aref pos 0))
				(Y0 (aref pos 1))
				(Z0 (aref pos 2))
				(R s-radius)
				(Xc (aref s-center 0))
				(Yc (aref s-center 1))
				(Zc (aref s-center 2))
				
				(dx (- X0 Xc))
				(dy (- Y0 Yc))
				(dz (- Z0 Zc))

				(B (* 2 (+ 	(* Xd dx)
										(* Yd dy)
										(* Zd dz))))
				(C (- (+ (* dx dx) (* dy dy) (* dz dz)) (* R R)))
				(discrim (- (* B B) (* 4 C)))
				)
				; in
				(if (< discrim 0) nil
					(let* (
						(discrim2 (sqrt discrim))					
						(t0 (/ (- (- B) discrim2) 2))
						(t1 (/ (+ (- B) discrim2) 2))
						(tt 0)
						) 
						;in	
						(if (>= t0 0) (setq tt t0) (setq tt t1))
						(make-array 3 :initial-contents 
							(list (+ X0 (* Xd tt)) (+ Y0 (* Yd tt)) (+ Z0 (* Zd tt)))
						)
					)
				)
			)
		) ;end intersection

		; shader
		(lambda (pos bounce)
			; calculate the lighting
			; use a simple diffuse phong model
			; i.e.
		
			; colour = SUM_ALL_LIGHTS(s-diffuse*light-diff*(L.N) + s-specular*light-specular*(R.E)^s-shininess) + Ka*amb			
			; where L is the light direction, N is the surface normal, R is the reflection vector (for the light) and E is the eye vector
			(let ((col *black*) (N (direction s-center pos)) (E (direction pos *origin*)))
				(dolist (light light-list)
					(let* (									
									(L (direction pos (light-pos light))) 
									(LdotN (max 0 (dot L N)))
									(R (subv (smulv N (* LdotN 2)) L))
									(RE (subv (smulv N (* (max (dot E N) 0) 2)) E))

									; diff
									(ldif (light-diffuse light))
									
									(ddark (mulc ldif s-dark))
									(dlight (mulc ldif s-light))

									; spec
									(lspec (light-specular light))
									(RdotE (max (dot R E) 0))
									(spec (mulc lspec s-specular))

									; uv coords									
									(x (- (aref pos 0) (aref s-center 0)))
									(y (- (aref pos 1) (aref s-center 1)))
									(phi (acos y))
									(theta (acos (/ x (sin phi)))) 
									(black-sq (mod (+ (floor (* phi 5)) (floor (* theta 5))) 2))
								)
						;in
						
						(setq col (addc col (smulc (if (eq black-sq 1) ddark dlight) LdotN)))
						(when (> RdotE 0.001)
							(setq shine (expt RdotE s-shininess))
							(setq col (addc col (smulc spec shine)))
						)
						
						(when (< bounce *max-bounces*)
							(let ((ref (cast-ray (addv pos (smulv N 0.01)) RE bounce)))
								(setq col (addc (smulc ref s-reflection) (smulc col (- 1 s-reflection))))
							)
						)

					)
				)
				(setq col (addc col (mulc *ambient* s-ambient)))
				col
			)
		) ;end shader
		)
	)
)

; a simple plane lying in x-z with normal y
(let (
			(s-dark (colour 230 230 230)) 
			(s-light (colour 20 20 20))
			(s-ambient *white*) 
			(s-specular *white*) 
			(s-shininess 1) 
			(s-ypos -1.05)
			(s-reflection 0.2)
		)
	(setq plane (list 
		;intersect function
		(lambda (pos dir)
			(let* (
				(Xd (aref dir 0))
				(Yd (aref dir 1))
				(Zd (aref dir 2))
				(X0 (aref pos 0))
				(Y0 (aref pos 1))
				(Z0 (aref pos 2))

				(numer (- s-ypos Y0))
				(denom Yd))
				;in

				(if (eq denom 0.) nil
				;else
					(let ((tt (/ numer denom)))
						(if (< tt 0) nil
						;else
							(vec (+ X0 (* tt Xd)) s-ypos (+ Z0 (* tt Zd)))
						)
					)
				)
			)
		) ;end intersection

		(lambda (pos bounce)
			; calculate the lighting
			; use a simple diffuse phong model
			; i.e.
			
			; colour = SUM_ALL_LIGHTS(s-diffuse*light-diff*(L.N) + s-specular*light-specular*(R.E)^s-shininess) + Ka*amb			
			; where L is the light direction, N is the surface normal, R is the reflection vector (for the light) and E is the eye vector
			(let ((col *black*) (N (vec 0 1 0)) (E (direction pos *origin*)))
				(dolist (light light-list)
					(let* (									
									(L (direction pos (light-pos light))) 
									(LdotN (max 0 (dot L N)))
									
									(R (subv (smulv N (* LdotN 2)) L))
									(RE (subv (smulv N (* (max (dot E N) 0) 2)) E))

									; diff
									(ldif (light-diffuse light))
									
									(diffblack (mulc ldif s-dark))
									(difflight (mulc ldif s-light))

									; spec
									(lspec (light-specular light))
									(RdotE (max (dot R E) 0))
									(spec (mulc lspec s-specular))

									; square
									(black-sq (mod (+ (floor (aref pos 0)) (floor (aref pos 2))) 2))
								)
						;in
						(if (intersect (addv pos (smulv N 0.01)) L) (setq col *black*)
						;else
						(progn
						
							(setq col (addc col (smulc (if (= black-sq 1) diffblack difflight) LdotN)))
							(when (> RdotE 0.001)
								(setq shine (expt RdotE s-shininess))
								(setq col (addc col (smulc spec shine)))
							)
							
							(when (< bounce *max-bounces*)
								(let ((ref (cast-ray (vec (aref pos 0) (+ s-ypos 0.001) (aref pos 2)) RE bounce)))
									(setq col (addc (smulc ref s-reflection) (smulc col (- 1 s-reflection))))
								)
							)
						)
						)
					)
				)
				(setq col (addc col (mulc *ambient* s-ambient)))
				col
			)
		) ;end shader
		)
	)
)
	
(push sphere object-list)
(push plane object-list)
(push (make-light :pos (vec -5 20 20)) light-list)

; sample invocation 
(time (render 300 300 "test.ppm"))

